#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include "HDCDtables.c"


#define ARITHMETIC_BIAS 0.49999

#if !defined(min)
#define min(x, y) ((x) <= (y) ? (x) : (y))
#endif

#define GAIN_APPLY(sample, gain) (sample) *= (gain)


    /* structures */ 

struct wav_header {
        char ckID[ 4 ];
        long cksize;                    /* expected to be 4 bytes */
        char WAVEID[ 4 ];
};

struct wav_format {
        char ckID[ 4 ];
        long cksize;
        short wFormatTag;
        short nChannels;
        long nSamplesPerSec;
        long nAvgBytesPerSec;
        short nBlockAlign;
        short wBitsPerSample;
};

struct wav_data {
        char ckID[ 4 ];
        long cksize;
};


typedef struct
{
    uint64_t window;
    unsigned char readahead, arg, control;
    int running_gain;
    unsigned sustain, sustain_reset;
} hdcd_state_t;


   /* global variables */

long int code_counter, code_counterA, code_counterB, code_counterC;


    /* proc declarations */


static int *read_wav( char*, long*, short*, long* );
static void write_wav( char*, int*, short, int, long  );
void hdcd_reset(hdcd_state_t*, unsigned int );
void hdcd_process(hdcd_state_t*, int*, int, int );
static int integrate(hdcd_state_t*, int*, int const*, int, int );
static int hdcd_scan(hdcd_state_t*, int const*, int, int );
static int hdcd_envelope(int*, int, int, int, int, int );



int main( int argc, char **argv )
{
        hdcd_state_t    decoder[ 2 ];
        long rate;
        short chan;
        int *samples;
        long count;
        
        code_counter=0; code_counterA=0; code_counterB=0; code_counterC=0;

        if ( argc != 3 ) {
                fprintf( stderr, "usage: hdcd_decode infile outfile\n" );
                exit( -1 );
        }

        samples = read_wav( argv[ 1 ], &count, &chan, &rate );

        hdcd_reset( &decoder[ 0 ], rate );
        hdcd_reset( &decoder[ 1 ], rate );

        printf( "chan1\n" );
        hdcd_process( &decoder[ 0 ], samples, count, chan );
        printf( "chan2\n" );
        hdcd_process( &decoder[ 0 ], samples + 1, count, chan );

        write_wav( argv[ 2 ], samples, chan, count, rate );
        
        
        printf("Code counting results\n");
        printf("Total = %ld  A = %ld    B = %ld     C= %ld\n",code_counter,code_counterA,code_counterB,code_counterC);

        return 0;
}


static int *read_wav( char *name, long *count, short *chan, long *rate ) 
{
        FILE *fp;
        struct wav_header header;
        struct wav_format format;
        struct wav_data data;
        int *samples;
        int i;

        fp = fopen( name, "r" );

        if ( !fp ) {
                fprintf( stderr, "hdcd_decode: failed to open '%s' for reading\n", name );
                exit( -2 );
        }

        if ( fread( &header, sizeof( header ), 1, fp ) != 1 ) {
                fprintf( stderr, "hdcd_decode: failed to read header\n" );
                exit( -3 );
        }

        if ( memcmp( header.ckID, "RIFF", 4 ) || memcmp( header.WAVEID, "WAVE", 4 )) {
                fprintf( stderr, "hdcd_decode: input file not a WAV\n" );
                exit( -3 );
        }

        if ( fread( &format, sizeof( format ), 1, fp ) != 1 ) {
                fprintf( stderr, "hdcd_decode: failed to read format\n" );
                exit( -4 );
        }

        if ( memcmp( format.ckID, "fmt ", 4 )) {
                fprintf( stderr, "hdcd_decode: input file not a WAV\n" );
                exit( -3 );
        }

        if ( format.wFormatTag != 0x0001 ) {
                fprintf( stderr, "hdcd_decode: input file not PCM\n" );
                exit( -4 );
        }

        if ( format.nChannels != 2 ) {
                fprintf( stderr, "hdcd_decode: input file not two chan WAV\n" );
                exit( -5 );
        }

        *rate = format.nSamplesPerSec;
        *chan = format.nChannels;

        if ( format.nBlockAlign != 4 ) {
                fprintf( stderr, "hdcd_decode: input file not 16 bit WAV\n" );
                exit( -6 );
        }

        if ( format.wBitsPerSample != 16 ) {
                fprintf( stderr, "hdcd_decode: input file not 16 bit WAV\n" );
                exit( -7 );
        }

        /*
         * Look for data section
         */

        do {
                if ( fread( &data, sizeof( data ), 1, fp ) != 1 ) {
                        fprintf( stderr, "hdcd_decode: failed to read data\n" );
                        exit( -8 );
                }
                if ( memcmp( data.ckID, "data", 4 ) == 0 ) {
                        break;
                }
                else {
                        fseek( fp, data.cksize, SEEK_CUR );
                }
        }
        while( 1 );

        if ( memcmp( data.ckID, "data", 4 )) {
                fprintf( stderr, "hdcd_decode: input file not a WAV\n" );
                exit( -8 );
        }

        *count = data.cksize / format.nBlockAlign;

        samples = malloc( sizeof( int ) * *chan * *count );
        if ( !samples ) {
                fprintf( stderr, "hdcd_decode: malloc failed\n" );
                exit( -9 );
        }

        printf( "hdcd_decode: decoding input file: chan's = %d, rate = %ld, samples = %ld\n", *chan, *rate, *count );

        for ( i = 0; i < *chan * *count; i ++ ) {
                short s;
                if ( fread( &s, 2, 1, fp ) != 1 ) {
                        fprintf( stderr, "hdcd_decode: file read failed\n" );
                        exit( -10 );
                }
                samples[ i ] = s;
        }

        fclose( fp );

        printf( "hdcd_decode: wav file reading complete\n" );

        return samples;
}

static void write_wav( char *name, int *samples, short chan, int count, long rate )
{
        FILE *fp;
        struct wav_header header;
        struct wav_format format;
        struct wav_data data;
        int i;

        fp = fopen( name, "w" );

        if ( !fp ) {
                fprintf( stderr, "hdcd_decode: failed to open '%s' for writing\n", name );
                exit( -20 );
        }

        memcpy( header.ckID, "RIFF", 4 );
        memcpy( header.WAVEID, "WAVE", 4 );
        header.cksize = 4 + sizeof( format ) + sizeof( data ) + ( count * chan * 2 );

        memcpy( format.ckID, "fmt ", 4 );
        format.cksize = 16;
        format.wFormatTag = 0x0001;     /* PCM */
        format.nChannels = 2;
        format.nSamplesPerSec = rate;
        format.nAvgBytesPerSec = rate * 6;
        format.nBlockAlign = 6;
        format.wBitsPerSample = 24;

        memcpy( data.ckID, "data", 4 );
        data.cksize = count * chan * 3;

        if ( fwrite( &header, sizeof( header ), 1, fp ) != 1 ) {
                fprintf( stderr, "hdcd_decode: file write failed\n" );
                exit( -21 );
        }
        if ( fwrite( &format, sizeof( format ), 1, fp ) != 1 ) {
                fprintf( stderr, "hdcd_decode: file write failed\n" );
                exit( -22 );
        }
        if ( fwrite( &data, sizeof( data ), 1, fp ) != 1 ) {
                fprintf( stderr, "hdcd_decode: file write failed\n" );
                exit( -23 );
        }

        for ( i = 0; i < count * chan; i ++ ) {
                int s;
                unsigned char x[ 3 ];

                /*
                 *  nasty endian fixed hack
                 */
                s = samples[ i ];
                s >>= 8;
                x[ 0 ] = s;
                s >>= 8;
                x[ 1 ] = s;
                s >>= 8;
                x[ 2 ] = s;

                if ( fwrite( x, 3, 1, fp ) != 1 ) {
                        fprintf( stderr, "hdcd_decode: file write failed\n" );
                        exit( -24 );
                }
        }

        printf( "hdcd_decode: wav file writing complete\n" );

        fclose( fp );
}


void hdcd_reset(hdcd_state_t *state, unsigned rate)
{
    if (gaintabup[15 << 7] == 0.0)
    {
        int i;
        for (i = 0; i < (1 << 8); i++)
            readaheadtab[i] = 32;
        for (i = 31; i > 0; i--)
        {
            uint32_t hash1 = 0x7e0fa005, hash2 = 0x7e0fa006;
            uint32_t step = 0x80000000u;
            hash1 >>= i;
            hash2 >>= i;
            step >>= i;
            step <<= 1;
            hash1 &= ~(-1 << 8);
            hash2 &= ~(-1 << 8);
            step &= ~(-1 << 8);
            do {
                readaheadtab[hash1] = i;
                readaheadtab[hash2] = i;
                hash1 += step;
                hash2 += step;
            } while (step && hash1 < (1 << 8));
        }
#if defined(ARITHMETIC)
        double acc = 1.0;
        for (i = 0; i < 15 << 7; i++)
        {
            if ((i & 127) == 0)
                acc = (float [] ){ 1.0, 0.94406128, 0.89126587, 0.84140015, 0.79434204, 0.74990845, 0.70794678, 0.66833496, 0.63095093, 0.59567261, 0.56234741, 0.53088379, 0.50119019, 0.47314453, 0.44668579, 0.42169189 }[i >> 7];
            else
                acc *= 0.99955034255981445;
            gaintab[i] = (int)(acc * 4096 + ARITHMETIC_BIAS) / 4096.0;
        }
        int j;
        for (j = 0; j < 8; j++)
            for (i = (15 << 7) - j; i >= 0; i -= 8)
            {
                if ((i & 120) == 0)
                    acc = gaintab[i];
                else
                    acc *= 1.00360429286956809;
                gaintabup[i] = (int)(acc * 4096 + ARITHMETIC_BIAS) / 4096.0;
            }
#else
        for (i = 0; i <= 15 << 7; i++)
            gaintabup[i] = gaintab[i] = (float)pow(10.0, -0.1 / 512 * i);
#endif
    }
    state->window = 0;
    state->readahead = 32;
    state->arg = 0;
    state->control = 0;

    state->running_gain = 0;

    state->sustain = 0;
    state->sustain_reset = rate * 10;
}

static int integrate(hdcd_state_t *state, int *flag, int const *samples, int count, int stride)
{
    uint32_t bits = 0;
    int result = min(state->readahead, count);
    int i;
    *flag = 0;
    for (i = result - 1; i >= 0; i--)
    {
        bits |= (*samples & 1) << i; /* might be better as a conditional? */
        samples += stride;
    }
    state->window = (state->window << result) | bits;
    state->readahead -= result;
    if (state->readahead > 0)
        return result;
    bits = (state->window ^ state->window >> 5 ^ state->window >> 23) & 0xffffffffu;
    if (state->arg)
    {
        if ((bits & 0xffffffc8) == 0x0fa00500)
        {
            state->control = (bits & 255) + (bits & 7);
            *flag = 1;
            code_counter++;
            code_counterA++;
        }
        if (((bits ^ (~bits >> 8 & 255)) & 0xffff00ff) == 0xa0060000)
        {
            state->control = bits >> 8 & 255;
            *flag = 1;
            code_counter++;
            code_counterB++;
        }
        state->arg = 0;
    }
    if (bits == 0x7e0fa005 || bits == 0x7e0fa006)
    {
        state->readahead = (bits & 3) * 8;
        state->arg = 1;
        code_counter++;
        code_counterC++;
    }
    else
    {
        if (bits)
            state->readahead = readaheadtab[bits & ~(-1 << 8)];
        else
            state->readahead = 31; /* ffwd over digisilence */
    }
    return result;
}

static int hdcd_scan(hdcd_state_t *state, int const *samples, int max, int stride)
{
    int result;
    int control;
    if (state->sustain > 0)
    {
        if (state->sustain <= max)
        {
            state->control = 0;
            max = state->sustain;
        }
        state->sustain -= max;
    }
	result = 0;
    while (result < max)
    {
        int flag;
        int consumed = integrate(state, &flag, samples, max - result, stride);
        result += consumed;
        if (flag > 0)
        {
            state->sustain = state->sustain_reset;
            break;
        }
        samples += consumed * stride;
    }
    return result;
}

static int hdcd_envelope(int *samples, int count, int stride, int gain, int target_gain, int extend)
{
    int i;

#if defined(ENVELOPE_ONLY)
        for (i = 0; i < count; i++)
            samples[i * stride] = 0x40000000;
#else
        if (extend)
        for (i = 0; i < count; i++)
        {
            int sample = samples[i * stride];
            int asample = abs(sample) - 0x5981;
            if (asample >= 0)
                sample = sample >= 0 ? peaktab[asample] : -peaktab[asample];
            else
                sample <<= 15;
            samples[i * stride] = sample;
        }
    else
        for (i = 0; i < count; i++)
            samples[i * stride] <<= 15;
#endif

    if (gain <= target_gain)
    {
        int len = min(count, target_gain - gain);
        /* attenuate slowly */
        for (i = 0; i < len; i++)
        {
            GAIN_APPLY(*samples, gaintab[++gain]);
            samples += stride;
        }
        count -= len;
    }
    else
    {
        int len = min(count, (gain - target_gain) >> 3);
        /* amplify quickly */
        for (i = 0; i < len; i++)
        {
            gain -= 8;
            GAIN_APPLY(*samples, gaintabup[gain]);
            samples += stride;
        }
        if (gain - 8 < target_gain) gain = target_gain;
        count -= len;
    }

    /* hold a steady level */
    if (gain == 0)
    {
        if (count > 0)
            samples += count * stride;
    }
    else
        while (--count >= 0)
        {
            GAIN_APPLY(*samples, gaintab[gain]);
            samples += stride;
        }

    return gain;
}

void hdcd_process(hdcd_state_t *state, int *samples, int count, int stride)
{
    int gain = state->running_gain;
    int peak_extend = (state->control & 16);
    int target_gain = (state->control & 15) << 7;
    int lead = 0;

    while (count > lead)
    {
        int run = hdcd_scan(state, samples + lead * stride, count - lead, stride) + lead;
        int envelope_run;

        envelope_run = run - 1;
        gain = hdcd_envelope(samples, envelope_run, stride, gain, target_gain, peak_extend);
        samples += envelope_run * stride;
        count -= envelope_run;
        lead = run - envelope_run;
        peak_extend = (state->control & 16);
        target_gain = (state->control & 15) << 7;
    }
    if (lead > 0)
        gain = hdcd_envelope(samples, lead, stride, gain, target_gain, peak_extend);
    state->running_gain = gain;
}
